module newPR_mod
  use sep 
  use igrad1
  implicit none
  integer     ,private   :: nx,nz,nw,nvx,nvz
  real        ,private   :: mxint,eps,mnint
  real        ,private   :: dtau,dx,dz,dw,ox,oz,ow,dvx,dvz,ovx,ovz
  logical     ,private   :: fromSP
contains
  subroutine newPR_mod_init(nz_in,nx_in,nw_in,dz_in,dx_in,dw_in,oz_in,ox_in,ow_in,dtau_in,mxint_in,fromSP_in,eps_in,mnint_in,nvz_in,nvx_in,dvz_in,dvx_in,ovz_in,ovx_in)
    integer        :: nz_in,nx_in,nw_in,nvz_in,nvx_in
    real           :: mxint_in,eps_in,mnint_in,dvz_in,dvx_in,ovz_in,ovx_in
    real           :: dtau_in,dz_in,dx_in,dw_in,ox_in,oz_in,ow_in
    logical        :: fromSP_in
    nz=nz_in; nx=nx_in; nw=nw_in; dz=dz_in; dtau=dtau_in; dx=dx_in
    mnint=mnint_in;nvx=nvx_in;nvz=nvz_in;dvx=dvx_in;dvz=dvz_in
    ovx=ovx_in;ovz=ovz_in
    oz=oz_in; ox=ox_in; ow=ow_in; dw=dw_in; mxint=mxint_in
    fromSP=fromSP_in;eps=eps_in
  end subroutine newPR_mod_init
  !####################################################################
  subroutine get_gradx(adj,add,data,PoverP_dx)
    integer                                   :: iz,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(KIND=8)   ,dimension(:,:)            :: PoverP_dx
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate(ctemp(nvx),rtemp(nvx),ctmp(nvx) )
    do iz=1,nvz
       stat=igrad1_lop(adj,add,real( data(iz,:)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(iz,:)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dx(iz,:)=aimag(ctmp/(data(iz,:)+eps))   
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradx
  !######################################################################
  subroutine get_gradz(adj,add,data,PoverP_dz)
    integer                                   :: ix,stat
    real           ,dimension(:)  ,allocatable:: ctemp,rtemp
    real(kind=8)   ,dimension(:,:)            :: PoverP_dz
    complex(kind=8),dimension(:)  ,allocatable:: ctmp
    complex        ,dimension(:,:)            :: data
    logical                                   :: adj,add
    allocate( ctemp(nvz),rtemp(nvz),ctmp(nvz) )
    do ix=1,nvx
       stat=igrad1_lop(adj,add,real( data(:,ix)),rtemp )
       stat=igrad1_lop(adj,add,aimag(data(:,ix)),ctemp )
       ctmp=cmplx(rtemp,ctemp)
       PoverP_dz(:,ix)=aimag(ctmp/(data(:,ix)+eps))
    end do
    deallocate(ctemp,rtemp,ctmp)
  end subroutine get_gradz
  !######################################################################
  subroutine Run_newPR_interp(vel,data,Rayout,adj,add,maxnum,x,z,xang,zang,it)
    integer                                :: ip,maxnum,xsgn,zsgn
    integer                                :: itau,it,iix,iiz
    integer     ,dimension(:),allocatable  :: xtmp,ztmp
    real(KIND=8)                           :: dx_step,dz_step,ds
    real(KIND=8)                           :: dx1,dz1,dx2,dz2,da,db,dc,dd,wa,wb,wc,wd
    real(KIND=8),dimension(:)              :: xang,zang,x,z
    real(KIND=8),dimension(:,:)            :: vel
    real(KIND=8),dimension(:,:),allocatable:: PoverP_dx,PoverP_dz
    real(KIND=8),dimension(:,:),allocatable:: xray,zray
    complex,     dimension(:,:)            :: Rayout
    complex,     dimension(:,:)            :: data
    logical                                :: adj,add  
    allocate( PoverP_dx(nvz,nvx),PoverP_dz(nvz,nvx))
    allocate( xtmp(1), ztmp(1) )
    allocate( xray(maxnum,nx),zray(maxnum,nx))
    xray(1,1:nx)=xang;  zray(1,1:nx)=zang
    call get_gradx(adj,add,data,PoverP_dx)
    call get_gradz(adj,add,data,PoverP_dz)
write(0,*) PoverP_dz(1,:)
    do itau=2, maxnum
!       write(0,*) 'Working on step ',itau, maxnum
       do ip=1, nx

          xtmp=minloc(abs(x-xray(itau-1,ip)));  iix=xtmp(1)
          ztmp=minloc(abs(z-zray(itau-1,ip)));  iiz=ztmp(1)

          dx1=dble( x(iix  )-xray(itau-1,ip))
          dz1=dble( z(iiz  )-zray(itau-1,ip))
          if (iix .gt. 1 .and. iix .lt. nvx) then
             if (dx1 .gt. 0.) then
               dx2=dble(x(iix-1)-xray(itau-1,ip))
               xsgn=-1
             else
               dx2=dble(x(iix+1)-xray(itau-1,ip))
               xsgn=1
             end if
          else
            xsgn=0
            dx2=dble(x(iix))
          end if
          if (iiz .gt. 1 .and. iiz .lt. nvz) then
            if (dz1 .gt. 0.) then
              dz2=dble(z(iiz-1)-zray(itau-1,ip))
              zsgn=-1
            else
              dz2=dble(z(iiz+1)-zray(itau-1,ip))
              zsgn=1
            end if
          else
            dz2=dble(z(iiz))
            zsgn=0
          end if
          da=dsqrt(dx1**2+dz1**2)
          db=dsqrt(dx1**2+dz2**2)
          dc=dsqrt(dx2**2+dz1**2)
          dd=dsqrt(dx2**2+dz2**2)
          wa=dble(1/(1+da/db+da/dc+da/dd))
          wb=((da/db)*wa)
          wc=((da/dc)*wa)
          wd=((da/dd)*wa)
          if (dx1 .eq. 0. .and. dz1 .eq.0) then
             wa=1.
             wb=0.;wc=0.;wd=0.
          end if
          ds=dble(dtau)*(wa*vel(iiz,iix)+wb*vel(iiz+zsgn,iix)+wc*vel(iiz,iix+xsgn)&
             +wd*vel(iiz+zsgn,iix+xsgn))

          if (iix .ge. nx-2*ceiling(abs(ds/dx))) then
             iix= nx-2*ceiling(abs(ds/dx))
          else if (iix .le. 2*ceiling(abs(ds/dx)) ) then
             iix= 2*ceiling(abs(ds/dx))
          else if (iiz .ge. nz-2*ceiling(abs(ds/dz)) )  then
             iiz= nz-2*ceiling(abs(ds/dz))
          end if

          !! Add onto existing rays
          xray(itau,ip)=xray(itau-1,ip)-ds*PoverP_dx(iiz,iix)/&
            sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)
          zray(itau,ip)=zray(itau-1,ip)-ds*PoverP_dz(iiz,iix)/&
            sqrt(PoverP_dx(iiz,iix)**2+PoverP_dz(iiz,iix)**2+eps)

       end do
    end do
    it=nx
    Rayout(1:maxnum,1:nx)=cmplx(xray,zray) 
    deallocate(PoverP_dx,PoverP_dz,xray,zray)
  end subroutine Run_newPR_interp
 !######################################################################
  subroutine write_on_baby(Rayout,maxnum,it,endang,begang)
    complex,  dimension(:,:)   :: Rayout
    integer                    :: it,maxnum
    real                       :: endang,begang
  call to_history("n1",maxnum          ) 
  call to_history("n2",nx              )  
  call to_history("d1",dtau            ) 
  call to_history("d2",dx              )  
  call to_history("o1",0.              ) 
  call to_history("o2",ox              )  
  call hclose()
  call sep_write(Rayout(1:maxnum,1:nx))
  end subroutine write_on_baby
 !######################################################################
end module newPR_mod
